#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
Collection of scripts to run gene finders, execute annotation pipelines, perform QC checks and generate summary statistics
"""


from jcvi.apps.base import dmain


if __name__ == "__main__":
    dmain(__file__)
