#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
Compilation of project specific scripts, used to execute specific analysis routines and generate publication-ready figures
"""

from jcvi.apps.base import dmain


if __name__ == "__main__":
    dmain(__file__)
