#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
Cluster of comparative genomics analysis methods: SynFind and QUOTA-ALIGN algorithms, synteny analysis, QC, etc. 
"""

from jcvi.apps.base import dmain


if __name__ == "__main__":
    dmain(__file__)
