#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
Suite of visualization tools for dot-plots, histograms, karytotypes, macro-/micro-synteny plots, seed counting using GRABSEEDS, etc. 
"""

from jcvi.apps.base import dmain


if __name__ == "__main__":
    dmain(__file__)
