#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
Set of scripts relating to variation studies such as imputation, phasing, SNP/CNV analysis, and other supporting routines
"""

from jcvi.apps.base import dmain


if __name__ == "__main__":
    dmain(__file__)
