#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
Assemblage of genome-assembly related scripts: ALLMAPS algorithm, scaffolding, k-mer analysis, QC, tool wrappers, etc. 
"""

from jcvi.apps.base import dmain


if __name__ == "__main__":
    dmain(__file__)
